/*
  PARRUN - directly receive into and execute from RAM via PARCP-USB
  by Daniel Illgen (insane.tscc.de)
  based upon PARCP by Petr Stehlik
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include "parcp-usb.h"

#define MAXSTRING 260
#define BLOCKSIZE (64*1024)
#define GAUGESIZ 20
#define HEADERSIZ 0x1c
#define ID_PRUN 0x5052554e
#define ID_OKAY 0x4f4b4159
#define ID_PROG 0x50524f47
#define ID_CDIR 0x43444952

int ctrl_c_flag = 0;
int file_open = 0;
FILE *file;

void ctrl_c_handler(int sig) {
	if (sig == SIGINT)
		ctrl_c_flag = 1;
}

void cleanup() {
	if (file_open) fclose(file);
	SET_INPUT;
	STROBE_HIGH;
	usb_exit();
}

void checkctrl_c() {
	if (ctrl_c_flag) {
		printf("Waiting for other side aborted by user\n");
		cleanup();
		exit(-1);
	}
}

void wait_before_read() {
	SET_INPUT;
	STROBE_LOW;
	while(IS_READY) checkctrl_c();
}

long read_long() {
	uint8_t inbuf[5];

	client_read_block(inbuf,4);
	return(inbuf[0]<<24)|(inbuf[1]<<16)|(inbuf[2]<<8)|(inbuf[3]);
}

void write_long(long x) {
	uint8_t outbuf[5];

	outbuf[0]=(x>>24)&0xff;
	outbuf[1]=(x>>16)&0xff;
	outbuf[2]=(x>>8)&0xff;
	outbuf[3]=x&0xff;
	client_write_block(outbuf,4);
}

int main(int argc, char *argv[]) {
	char filename[MAXSTRING]="",stdir[MAXSTRING]="";
	char usb_serial[MAXSTRING] = "";
	uint8_t buf[BLOCKSIZE];
	uint8_t hdr[HEADERSIZ];
	uint32_t dirlen,filelen,fileread,parread,btodo;
  	int	cnt,cntmax;

	switch (argc) {
		case 2: //only filename given
			strcpy(filename,argv[1]);
			break;
		case 3: //st-dir and filename
			strcpy(stdir,argv[1]);
			strcpy(filename,argv[2]);
			break;
		default:
			printf("usage: parrun [st-directory] file \n");
			exit(0);
			break;
	}

	signal(SIGINT,ctrl_c_handler);
	printf("running \"%s\" in dir \"%s\"\n",filename,stdir);

	if (usb_init(usb_serial) == 0) {
		fprintf(stderr, "USB init failed\n");
		return(-1);
	}

	SET_INPUT;
	STROBE_HIGH;

	if (!IS_READY) {
		printf("Cannot connect to Server on ATARI side - please start the server first\n");
		usb_exit();
		return(-1);
	}

	write_long(ID_PRUN);
	wait_before_read();
	parread=read_long();
	if (parread!=ID_OKAY) {
		printf("Other side is not PARRUN Server\n");
		usb_exit();
		return(-1);
	}

	if ((file=fopen(filename,"rb"))==NULL) {
		usb_exit();
		printf("Cannot open %s\n",filename);
		return(-1);
	}
	file_open = 1;
	fseek(file,0L,SEEK_END);
	filelen=ftell(file);
	rewind(file);

	fileread=fread(hdr,1,HEADERSIZ,file);
	filelen-=fileread;btodo=filelen;

	if ((btodo<0x1e)||(((hdr[0]<<8)|(hdr[1]))!=0x601a)) {
		printf("Not an ATARI Program: %s\n",filename);
		cleanup();
		return(-1);
	}

	if (strcmp(stdir,"")!=0) {
		dirlen=strlen(stdir)+1;
		if (dirlen<250) {
			write_long(ID_CDIR);
			write_long(dirlen);
			client_write_block((unsigned char *)stdir,dirlen);
			wait_before_read();
			parread=read_long();
			if (parread!=ID_OKAY) printf("Error changing directory: %X\n",parread);
		}
	}

	write_long(ID_PROG);
	client_write_block(hdr,HEADERSIZ);
	wait_before_read();
	parread=read_long();
	if (parread != ID_OKAY) {
			printf("program not executable on ATARI - not enough process memory\n");
			cleanup();
			return(-1);
	}

	while(btodo>0) {
		fileread=fread(buf,1,BLOCKSIZE,file);
		btodo-=fileread;
		if (!btodo) fileread=(fileread+3)&0xfffffc;
		write_long(fileread);
		wait_before_read();
		parread=read_long();

		cntmax=btodo*GAUGESIZ/filelen;
		printf("\rsending: [");
		for(cnt=0;cnt<(GAUGESIZ-cntmax);cnt++) putchar('#');
		for(cnt=0;cnt<cntmax;cnt++) putchar(':');
		putchar(']');
		fflush(stdout);

		if (parread != fileread) {
			printf("\ncommunication error: send %d, received: %d\n",fileread,parread);
			cleanup();
			return(-1);
		}

		client_write_block(buf,fileread);

		checkctrl_c();
	}
	printf("\n");
	write_long(0);
	cleanup();

	return(0);
}

/* vim: set ts=4: */
